#include "includes/common_functions.inc"

// * User input * //
uniform mediump sampler2D u_texture;
uniform mediump sampler2D u_texture1;
uniform mediump sampler2D u_texture2;
uniform mediump sampler2D u_texture3;

uniform highp float u_time;	
uniform highp vec4 u_screenSize;

// Light matrix in world space
uniform highp mat4 u_directionalLightMatrix; 
// Dir Color in RGB , intensity in A
uniform lowp vec4 u_directionalLightColor; 
// Ambient Color in RGB, intensity in A (contrast)
uniform lowp vec4 u_ambientLightColor;
	
// .x and .y = rim controls, .z = galaxy pattern size .w = color change timing
uniform lowp vec4 u_rims_and_galaxyscale;

// Special handling for 2D UI world...
uniform lowp vec4 u_uicontrol;
	
// Data IN FROM VERTEX SHADER
varying mediump vec3 v_normal;
varying mediump vec3 v_tangent;
varying mediump vec3 v_binormal;
varying mediump vec4 v_vert_color;
varying lowp vec2 v_texCoord0;
varying lowp vec3 v_camDir;

void main() {
	// NORMALIZE INPUT VALUES TO ENSURE COUNTER INTERPOLATION ISSUES
	mediump vec3 L = normalize(-u_directionalLightMatrix[2].xyz);
	mediump vec3 V = normalize(v_camDir);
	mediump vec3 N = normalize(v_normal);
	mediump vec3 B = normalize(v_binormal);
	mediump vec3 T = normalize(v_tangent);
	
	// Metalness, Roughness, Wetness -----------------------------------------------------------------------------
	mediump vec3 mrw = v_vert_color.rgb;
	
	// ALBEDO MAP  -----------------------------------------------------------------------------
	lowp vec4 albedo = texture2D(u_texture, v_texCoord0);
	albedo.rgb = pow(albedo.rgb, vec3(2.2, 2.2, 2.2));

	// NORMAL MAP  -----------------------------------------------------------------------------
	mediump vec3 normalmap = texture2D(u_texture1, v_texCoord0).xyz * 2.0 - 1.0;
	N = normalize(mat3(T, B, N) * normalmap);
		
	// DOT PRODUCTS + REFLECTION VECTORS
	lowp float NdotV = dot(N, V);
	lowp float NdotL = (dot(N, L)+ 1.0) * 0.5;   // 1/2 LAMBERTIAN LIGHTING MODEL
	mediump vec3 R = reflect(-V, N*vec3(1.0 , u_uicontrol.y, 1.0));
	lowp float NormalUp = u_uicontrol.z + u_uicontrol.w * (N.y + 1.0)*0.5;
	
	// LIGHTING CALCLUATIONS
	lowp vec3 dir_light = u_directionalLightColor.rgb * NdotL * u_directionalLightColor.a ;
	lowp vec3 ambient_light = mix(u_ambientLightColor.rgb * vec3(0.8, 0.25, 0.1), u_ambientLightColor.rgb, NormalUp) * u_ambientLightColor.a;
		
	// GALAXY COLOR TINT
	lowp float average = (albedo.r + albedo.g + albedo.b) * 0.333;
	highp float speed = u_time *u_rims_and_galaxyscale.w;
	mediump vec3 colortint= vec3(sin(speed+average),sin(speed+degrees(45.0*0.5)),sin(speed+degrees(45.0)));
	colortint=abs(colortint*colortint*colortint);	

	// GALAXY PATTERNS 
	highp float scrolltime = fract(u_time*0.003);
	vec2 uv =(gl_FragCoord.xy / u_screenSize.y)*u_rims_and_galaxyscale.z;
	
	highp vec3 pattern = vec3(0.0);
	highp vec3 star=vec3(0.0);
	highp vec3 galaxy=vec3(0.0);
	
	star.r = texture2D(u_texture2, 2.0*fract(uv + vec2(1.0-scrolltime*7.0, scrolltime*5.0))).b;
	star.g = texture2D(u_texture2, 3.0*fract(uv + vec2(1.0-scrolltime*4.0, scrolltime))).b;
	
	star = smoothstep(0.3,0.7, clamp(star, 0.0, 0.9));
	star +=star;
	star.b=texture2D(u_texture2,2.0*fract( uv + vec2(scrolltime*17.0, 0.0))).r;
		
	pattern.r = texture2D(u_texture2,fract( uv + vec2(1.0-scrolltime*7.0, 0.0))).r;	
	pattern.g = texture2D(u_texture2, 3.0*fract(uv + vec2(1.0-scrolltime*5.0, 0.0))).g;
	pattern.b = texture2D(u_texture2, 2.0*fract(uv + vec2(scrolltime*8.0, 0.0))).g;
	
	//Putting it all together	
	galaxy.rgb = vec3(colortint.r, colortint.g, 1.0)*(pattern.g+pattern.b);
	galaxy.rgb = galaxy.rgb*vec3(1.0-colortint.r, 1.0-colortint.g, 1.0)+vec3(0.2,0.2,0.8)*pattern.b*0.8;
	galaxy.rgb = galaxy.rgb+(colortint*vec3(0.6,0.2,0.4)*(pattern.r*0.7));
	galaxy.rgb = galaxy.rgb+(vec3(0.7,0.2,0.7)*(pattern.r*pattern.b));	
	//adding stars
	galaxy.rgb += clamp(colortint, 0.5,0.9) *star.g*star.b + vec3(0.8,1.0,0.57)*star.r*star.b;
	//masking between albedo and galaxy
	albedo.rgb=mix(galaxy.rgb,albedo.rgb,albedo.a);
	
	// SPEC ------------------  -----------------------------------------------------------------------------
	mediump vec3 specColour = mix(vec3(0.03, 0.03, 0.03), albedo.rgb * 2.0, mrw.r);	
	highp vec2 sphericalCoords = vert_CalculateSphericalCoords(mrw.g, R);
	
	//ENV SAMPLING
	lowp vec4 envMapTex = texture2D(u_texture3, sphericalCoords);
	envMapTex = pow(envMapTex, vec4(2.2, 2.2, 2.2, 2.2));
	
	//Function CalculateEnvSpec accepts parameters (vec4 envMap, vec3 mrw, float NdotV, vec3 amb_color, vec3 spec_color)
	//Returns vec4 envSpec, env map colors in rgb, spec coefficient in a.
	mediump vec4 envSpec = CalculateEnvSpec(envMapTex, mrw, NdotV, ambient_light, specColour);
	mediump vec3 spec_BRDF = CalculateBRDF(NdotV, mrw.g, specColour, dir_light, envSpec.a, v_vert_color.a );	

	// RIM LIGHTING   ---------------------------------------------------------------------------------
	lowp vec3 rimLight = mix(dir_light.rgb ,albedo.rgb , NdotL);
	rimLight *= pow(max(0.0001, 1.0 - NdotV), u_rims_and_galaxyscale.y) * (u_uicontrol.y * N.y + 1.0) * u_rims_and_galaxyscale.x * v_vert_color.a;
	
	// BUILD FRAGMENT OUTPUT  -------------------------------------------------------------------------
	mediump vec4 returnValue = vec4(albedo.rgb *(dir_light + ambient_light) , 1.0);
	returnValue.rgb += spec_BRDF.rgb * NormalUp;
	returnValue.rgb += envSpec.rgb;
	returnValue.rgb += rimLight;
	
	// MOVE BACK INTO sRGB COLOUR SPACE
	returnValue.rgb = pow(returnValue.rgb, vec3(1.0/2.2, 1.0/2.2, 1.0/2.2));
	gl_FragColor = returnValue;
}